<?php

/**
 * Notifier class has purpose of sending various notification to users. Primary
 * notification method is email
 *
 * @version 1.0
 * @author Ilija Studen <ilija.studen@gmail.com>
 */
class Notifier {

	/** Supported transports **/
	const MAIL_TRANSPORT_MAIL = 'mail()';
	const MAIL_TRANSPORT_SMTP = 'smtp';

	/** Secure connection values **/
	const SMTP_SECURE_CONNECTION_NO  = 'no';
	const SMTP_SECURE_CONNECTION_SSL = 'ssl';
	const SMTP_SECURE_CONNECTION_TLS = 'tls';

	/**
	 * Cached value of echange compatible config option
	 *
	 * @var boolean
	 */
	static public $exchange_compatible = null;

	function notifyAction($object, $action, $log_data) {
		if (!$object instanceof ProjectDataObject) {
			return;
		}
		$subscribers = $object->getSubscribers();
		if (!is_array($subscribers) || count($subscribers) == 0) return;
		if ($action == ApplicationLogs::ACTION_ADD) {
			if ($object instanceof ProjectMessage) {
				self::newMessage($object, $subscribers);
			} else if ($object instanceof Comment) {
				self::newObjectComment($object);
			} else {
				self::newObject($object, $subscribers);
			}
		} else if ($action == ApplicationLogs::ACTION_EDIT) {
			self::objectEdited($object, $subscribers);
		} else if ($action == ApplicationLogs::ACTION_TRASH) {
			self::objectDeleted($object, $subscribers);
		} else if ($action == ApplicationLogs::ACTION_CLOSE) {
			self::objectClosed($object, $subscribers);
		}
	}
	
	function newObject(ProjectDataObject $object, $people) {
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('object', $object);

		$recepients = array();
		foreach($people as $user) {
			if ($user->getId() != $object->getCreatedById()) {
				$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
			}
		} // foreach
		
		if (count($recepients) == 0) return;
		if (! $object->getCreatedBy() instanceof User) return;
		
		return self::sendEmail(
			$recepients,
			self::prepareEmailAddress($object->getCreatedBy()->getEmail(), $object->getCreatedByDisplayName()),
			lang('new notification ' . $object->getObjectTypeName(), $object->getObjectName()),
			tpl_fetch(get_template_path('new_object', 'notifier'))
		); // send
	}
	
	function shareObject(ProjectDataObject $object, $people) {
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('object', $object);

		$recepients = array();
		foreach($people as $user) {
			if ($user->getId() != $object->getCreatedById()) {
				$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
			}
		} // foreach
		
		if (count($recepients) == 0) return;
		if (! $object->getCreatedBy() instanceof User) return;
		
		return self::sendEmail(
			$recepients,
			self::prepareEmailAddress(logged_user()->getEmail(), logged_user()->getDisplayName()),
			lang('new share notification ' . $object->getObjectTypeName(), $object->getObjectName()),
			tpl_fetch(get_template_path('share_object', 'notifier'))
		); // send
	}
	
	function objectEdited(ProjectDataObject $object, $people) {
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('object', $object);

		$recepients = array();
		foreach($people as $user) {
			if ($user->getId() != $object->getUpdatedById()) {
				$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
			}
		} // foreach
		
		if (count($recepients) == 0) return;
		if (! $object->getUpdatedBy() instanceof User) return;
		
		return self::sendEmail(
			$recepients,
			self::prepareEmailAddress($object->getUpdatedBy()->getEmail(), $object->getUpdatedByDisplayName()),
			lang('modified notification ' . $object->getObjectTypeName(), $object->getObjectName()),
			tpl_fetch(get_template_path('object_edited', 'notifier'))
		); // send
	}
	
	function objectClosed(ProjectDataObject $object, $people) {
		if (!($object instanceof ProjectTask || $object instanceof ProjectMilestone)) {
			return;
		}
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if
		$closedBy = $object->getCompletedBy();
		if (!$closedBy instanceof User) return;
		tpl_assign('object', $object);
		tpl_assign('closedBy', $closedBy);

		$recepients = array();
		foreach($people as $user) {
			if ($user->getId() != $closedBy->getId()) {
				$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
			}
		} // foreach
		if (count($recepients) == 0) return;
		return self::sendEmail(
			$recepients,
			self::prepareEmailAddress($closedBy->getEmail(), $closedBy->getDisplayName()),
			lang('closed notification ' . $object->getObjectTypeName(), $object->getObjectName()),
			tpl_fetch(get_template_path('object_closed', 'notifier'))
		); // send
	}
	
	function objectDeleted(ProjectDataObject $object, $people) {
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('object', $object);

		$recepients = array();
		foreach($people as $user) {
			if ($user->getId() != $object->getTrashedById()) {
				$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
			}
		} // foreach

		$trashedBy = Users::findById($object->getTrashedById());
		if ($trashedBy instanceof User) {
			$displayName = $trashedBy->getDisplayName();
			$email = $trashedBy->getEmail();
		} else {
			$displayName = lang("n/a");
			$email = lang("n/a");
		}
		if (count($recepients) == 0) return;
		
		return self::sendEmail(
			$recepients,
			self::prepareEmailAddress($email, $displayName),
			lang('deleted notification ' . $object->getObjectTypeName(), $object->getObjectName()),
			tpl_fetch(get_template_path('object_deleted', 'notifier'))
		); // send
	}
	
	/**
	 * Reset password and send forgot password email to the user
	 *
	 * @param User $user
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	static function forgotPassword(User $user) {
		$administrator = owner_company()->getCreatedBy();

		$new_password = $user->resetPassword(true);
		tpl_assign('user', $user);
		tpl_assign('new_password', $new_password);
		
		if (! $administrator instanceof User) return;

		return self::sendEmail(
		self::prepareEmailAddress($user->getEmail(), $user->getDisplayName()),
		self::prepareEmailAddress($administrator->getEmail(), $administrator->getDisplayName()),
		lang('your password'),
		tpl_fetch(get_template_path('forgot_password', 'notifier'))
		); // send
	} // forgotPassword

	/**
	 * Send new account notification email to the user whose accout has been created
	 * (welcome message)
	 *
	 * @param User $user
	 * @param string $raw_password
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	static function newUserAccount(User $user, $raw_password) {
		tpl_assign('new_account', $user);
		tpl_assign('raw_password', $raw_password);

		if (! $user->getCreatedBy() instanceof User) return;
		
		return self::sendEmail(
		self::prepareEmailAddress($user->getEmail(), $user->getDisplayName()),
		self::prepareEmailAddress($user->getCreatedBy()->getEmail(), $user->getCreatedByDisplayName()),
		lang('your account created'),
		tpl_fetch(get_template_path('new_account', 'notifier'))
		); // send
	} // newUserAccount

	/**
	 * Send new message notification to the list of users ($people)
	 *
	 * @param ProjectMessage $message New message
	 * @param array $people
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	static function newMessage(ProjectMessage $message, $people) {
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('new_message', $message);

		$recepients = array();
		foreach($people as $user) {
			if ($user->getId() != $message->getCreatedById()) {
				$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
			}
		} // foreach

		if (count($recepients) == 0) return;
		if (! $message->getCreatedBy() instanceof User) return;
		
		return self::sendEmail(
		$recepients,
		self::prepareEmailAddress($message->getCreatedBy()->getEmail(), $message->getCreatedByDisplayName()),
		/*$message->getProject()->getName() .*/ ' - ' . lang('new message'),
		tpl_fetch(get_template_path('new_message', 'notifier'))
		); // send
	} // newMessage

	/**
	 * Send new task notification to the list of users ($people)
	 *
	 * @param ProjectTask $task New task
	 * @param array $people
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	static function newTask(ProjectTask $task, $people) {
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('new_task', $task);
		tpl_assign('is_new', true);

		$recepients = array();
		foreach($people as $user) {
			$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
		} // foreach

		if (! $task->getCreatedBy() instanceof User) return;
		
		return self::sendEmail(
		$recepients,
		self::prepareEmailAddress($task->getCreatedBy()->getEmail(), $task->getCreatedByDisplayName()),
		$task->getProject()->getName() . ' - ' . lang('new task'),
		tpl_fetch(get_template_path('new_task', 'notifier'))
		); // send
	} // newTask
	
	/**
	 * Send task changed notification to the list of users ($people)
	 *
	 * @param ProjectTask $task task
	 * @param array $people
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	static function taskChanged(ProjectTask $task, $people) {
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('new_task', $task);
		tpl_assign('is_new', false);

		$recepients = array();
		foreach($people as $user) {
			$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
		} // foreach

		if (! $task->getCreatedBy() instanceof User) return;
		
		return self::sendEmail(
		$recepients,
		self::prepareEmailAddress($task->getCreatedBy()->getEmail(), $task->getCreatedByDisplayName()),
		$task->getProject()->getName() . ' - ' . lang('task modified'),
		tpl_fetch(get_template_path('new_task', 'notifier'))
		); // send
	} // taskChanged
	
	/**
	 * Send task due notification to the list of users ($people)
	 *
	 * @param ProjectTask $task Due task
	 * @param array $people
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	static function objectReminder(ObjectReminder $reminder) {
		$object = $reminder->getObject();
		$context = $reminder->getContext();
		$type = $object->getObjectTypeName();
		$date = $object->getColumnValue($context);
		if ($reminder->getUserId() == 0) {
			$people = $object->getSubscribers();
		} else {
			$people = array($reminder->getUser());
		}
		Env::useHelper("format");
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('object', $object);
		tpl_assign('type', $type);
		tpl_assign('context', $context);
		tpl_assign('date', $date);

		$recepients = array();
		foreach($people as $user) {
			$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
		} // foreach

		if (!$object->getCreatedBy() instanceof User) return;
		
		return self::sendEmail(
			$recepients,
			self::prepareEmailAddress($object->getCreatedBy()->getEmail(), $object->getCreatedByDisplayName()),
			lang("$context $type reminder"),
			tpl_fetch(get_template_path('object_reminder', 'notifier'))
		); // send
	} // taskDue

	static function eventReminder(ProjectEvent $event, $people) {
		Env::useHelper("format");
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('event', $event);

		$recepients = array();
		foreach($people as $user) {
			$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
		} // foreach

		if (! $event->getCreatedBy() instanceof User) return;
		
		return self::sendEmail(
			$recepients,
			self::prepareEmailAddress($event->getCreatedBy()->getEmail(), $event->getCreatedByDisplayName()),
			$event->getProject()->getName() . ' - ' . lang('event reminder'),
			tpl_fetch(get_template_path('event_reminder', 'notifier'))
		); // send
	}
	
	/**
	 * Send event notification to the list of users ($people)
	 *
	 * @param ProjectEvent $event Event
	 * @param array $people
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	static function notifEvent(ProjectEvent $event, $people, $is_new = false) {
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('event', $event);
		tpl_assign('is_new', $is_new);

		$recepients = array();
		foreach($people as $user) {
			$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
		} // foreach
		
		if (! $event->getCreatedBy() instanceof User) return;

		return self::sendEmail(
			$recepients,
			self::prepareEmailAddress(($is_new ? $event->getCreatedBy()->getEmail() : $event->getUpdatedBy()->getEmail()),
			($is_new ? $event->getCreatedByDisplayName() : $event->getUpdatedByDisplayName())),
			$event->getProject()->getName() . ' - ' . ($is_new ? lang('new event notification') : lang('change event notification')) . ': ' . $event->getSubject(),
			tpl_fetch(get_template_path('event_notif', 'notifier'))
		); // send
	} // notifEvent
	
	 /** Send event notification to the list of users ($people)
	 *
	 * @param ProjectEvent $event Event
	 * @param array $people
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	static function notifEventDeletion($subject, $proj_name, $start_date, $people) {
		if(!is_array($people) || !count($people)) {
			return; // nothing here...
		} // if

		tpl_assign('is_deleted', true);
		tpl_assign('eventSubject', $subject);
		tpl_assign('projectName', $proj_name);
		tpl_assign('eventStart', $start_date);
		
		$recepients = array();
		foreach($people as $user) {
			$recepients[] = self::prepareEmailAddress($user->getEmail(), $user->getDisplayName());
		} // foreach

		return self::sendEmail(
			$recepients,
			self::prepareEmailAddress(logged_user()->getEmail(), logged_user()->getDisplayName()),
			$proj_name . ' - ' . lang('deleted event notification') . ': ' . $subject,
			tpl_fetch(get_template_path('event_notif', 'notifier'))
		); // send
	} // notifEvent
	
	/**
	 * Send new comment notification to message subscribers
	 *
	 * @param Comment $comment
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	static function newObjectComment(Comment $comment, $all_subscribers) {
		$object = $comment->getObject();
		if(!($object instanceof ApplicationDataObject)) {
			throw new Error('Invalid comment object');
		} // if

		$recepients = array();
		foreach($all_subscribers as $subscriber) {
			if($subscriber->getId() == $comment->getCreatedById()) {
				continue; // skip comment author
			} // if

			if($comment->isPrivate()) {
				if($subscriber->isMemberOfOwnerCompany()) {
					$recepients[] = self::prepareEmailAddress($subscriber->getEmail(), $subscriber->getDisplayName());
				} // if
			} else {
				$recepients[] = self::prepareEmailAddress($subscriber->getEmail(), $subscriber->getDisplayName());
			} // of
		} // foreach

		if(!count($recepients)) {
			return true; // no recepients
		} // if

		tpl_assign('new_comment', $comment);
		
		if (! $comment->getCreatedBy() instanceof User) return;

		return self::sendEmail($recepients,
				self::prepareEmailAddress($comment->getCreatedBy()->getEmail(), $comment->getCreatedByDisplayName()),
				lang('new comment') . ' - ' . $comment->getObject()->getObjectName(),
				tpl_fetch(get_template_path('new_comment', 'notifier'))
		); // send
	} // newObjectComment

	// ---------------------------------------------------
	//  Milestone
	// ---------------------------------------------------

	/**
	 * Milestone has been assigned to the user
	 *
	 * @param ProjectMilestone $milestone
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	function milestoneAssigned(ProjectMilestone $milestone) {
		if($milestone->isCompleted()) {
			return true; // milestone has been already completed...
		} // if
		if(!($milestone->getAssignedTo() instanceof User)) {
			return true; // not assigned to user
		} // if

		tpl_assign('milestone_assigned', $milestone);

		if (! $milestone->getCreatedBy() instanceof User) return;
		
		return self::sendEmail(
		self::prepareEmailAddress($milestone->getAssignedTo()->getEmail(), $milestone->getAssignedTo()->getDisplayName()),
		self::prepareEmailAddress($milestone->getCreatedBy()->getEmail(), $milestone->getCreatedByDisplayName()),
		lang('milestone assigned to you'),
		tpl_fetch(get_template_path('milestone_assigned', 'notifier'))
		); // send
	} // milestoneAssigned

	/**
	 * Task has been assigned to the user
	 *
	 * @param ProjectTask $task
	 * @return boolean
	 * @throws NotifierConnectionError
	 */
	function taskAssigned(ProjectTask $task) {
		if($task->isCompleted()) {
			return true; // task has been already completed...
		} // if
		if(!($task->getAssignedTo() instanceof User)) {
			return true; // not assigned to user
		} // if

		tpl_assign('task_assigned', $task);

		return self::sendEmail(
		self::prepareEmailAddress($task->getAssignedTo()->getEmail(), $task->getAssignedTo()->getDisplayName()),
		self::prepareEmailAddress($task->getUpdatedBy()->getEmail(), $task->getUpdatedByDisplayName()),
		lang('task assigned to you'),
		tpl_fetch(get_template_path('task_assigned', 'notifier'))
		); // send
	} // taskAssigned



	// ---------------------------------------------------
	//  Util functions
	// ---------------------------------------------------

	/**
	 * This function will prepare email address. It will return $name <$email> if both
	 * params are presend and we are not in exchange compatibility mode. In other case
	 * it will just return email
	 *
	 * @param string $email
	 * @param string $name
	 * @return string
	 */
	static function prepareEmailAddress($email, $name = null) {
		if(trim($name) && !self::getExchangeCompatible()) {
			return trim($name) . ' <' . trim($email) . '>';
		} else {
			return trim($email);
		} // if
	} // prepareEmailAddress

	/**
	 * Returns true if exchange compatible config option is set to true
	 *
	 * @param void
	 * @return boolean
	 */
	static function getExchangeCompatible() {
		if(is_null(self::$exchange_compatible)) {
			self::$exchange_compatible = config_option('exchange_compatible', false);
		} // if
		return self::$exchange_compatible;
	} // getExchangeCompatible

	/**
	 * Send an email using Swift (send commands)
	 *
	 * @param string to_address
	 * @param string from_address
	 * @param string subject
	 * @param string body, optional
	 * @param string content-type,optional
	 * @param string content-transfer-encoding,optional
	 * @return bool successful
	 */
	static function sendEmail($to, $from, $subject, $body = false, $type = 'text/plain', $encoding = '8bit') {
		Env::useLibrary('swift');

		$mailer = self::getMailer();
		if(!($mailer instanceof Swift)) {
			throw new NotifierConnectionError();
		} // if

		if (config_option("mail_transport", self::MAIL_TRANSPORT_MAIL) == self::MAIL_TRANSPORT_SMTP &&
				config_option("smtp_authenticate", false)) {
			$from = self::prepareEmailAddress(config_option("smtp_username", $from), $from);
		}
		$result = $mailer->send($to, $from, $subject, $body, $type, $encoding);
		$mailer->close();

		return $result;
	} // sendEmail

	/**
	 * This function will return SMTP connection. It will try to load options from
	 * config and if it fails it will use settings from php.ini
	 *
	 * @param void
	 * @return Swift
	 */
	static function getMailer() {
		$mail_transport_config = config_option('mail_transport', self::MAIL_TRANSPORT_MAIL);

		// Emulate mail() - use NativeMail
		if($mail_transport_config == self::MAIL_TRANSPORT_MAIL) {
			$mailer = new Swift(new Swift_Connection_NativeMail());
			return $mailer->isConnected() ? $mailer : null;

			// Use SMTP server
		} elseif($mail_transport_config == self::MAIL_TRANSPORT_SMTP) {

			// Load SMTP config
			$smtp_server = config_option('smtp_server');
			$smtp_port = config_option('smtp_port', 25);
			$smtp_secure_connection = config_option('smtp_secure_connection', self::SMTP_SECURE_CONNECTION_NO);
			$smtp_authenticate = config_option('smtp_authenticate', false);
			if($smtp_authenticate) {
				$smtp_username = config_option('smtp_username');
				$smtp_password = config_option('smtp_password');
			} // if

			switch($smtp_secure_connection) {
				case self::SMTP_SECURE_CONNECTION_SSL:
					$transport = SWIFT_SSL;
					break;
				case self::SMTP_SECURE_CONNECTION_TLS:
					$transport = SWIFT_TLS;
					break;
				default:
					$transport = SWIFT_OPEN;
			} // switch

			$mailer = new Swift(new Swift_Connection_SMTP($smtp_server, $smtp_port, $transport));
			if(!$mailer->isConnected()) {
				return null;
			} // if

			$mailer->setCharset('UTF-8');

			if($smtp_authenticate) {
				if($mailer->authenticate($smtp_username, $smtp_password)) {
					return $mailer;
				} else {
					return null;
				} // if
			} else {
				return $mailer;
			} // if

			// Somethings wrong here...
		} else {
			return null;
		} // if
	} // getMailer

	function sendReminders() {
		include_once "application/cron_functions.php";
		send_reminders();
	}
	
} // Notifier

?>
